#include <alloc.h>
#include <stdio.h>
#include <string.h>

#include <Application.h>
#include <Window.h>
#include <Region.h>
#include <Node.h>
#include <NodeInfo.h>
#include <Alert.h>
#include <Font.h>
#include <Directory.h>
#include <File.h>
#include <Path.h>
#include <Roster.h>
#include <String.h>

#include "Colors.h"
#include "CLVColumn.h"
#include "SafeTime.h"

#include "SeekerWindow.h"
#include "STreeView.h"

#include "SFileView.h"

SFileItem::SFileItem(SFileView *fileView, const SFile *file)
 : CLVListItem(0, false, false, 20.0),	// 20.0 is minimum row height
   iFileView(fileView),
   iFile(0),
   iTextOffset(0.0)
{
	// copy entry
	iFile=new SFile(file);
}

SFileItem::~SFileItem()
{
	
	if(iFile) delete iFile;
}

// custom paint
void SFileItem::DrawItemColumn(BView *owner, BRect itemColumnRect, 
								int32 columnIndex, bool complete)
{
	rgb_color color;
	bool selected=IsSelected();
	
	if(selected) {
		color=Black;
	} else {
		color=White;
	}
	
	owner->SetLowColor(color);
	owner->SetDrawingMode(B_OP_COPY);
	if((selected || complete) && (columnIndex==1)) {
		owner->SetHighColor(color);

		BRect rect=itemColumnRect;
		rect.left -=1.0;
		rect.top +=iTextOffset - ceil(iFontAttr.ascent);
		rect.bottom -=((iTextOffset - ceil(iFontAttr.ascent)) / 2.0) + 2.0;
		rect.right=rect.left + owner->StringWidth(iFile->Name()) + 3.0;

		owner->FillRect(rect);
	}
	
	BRegion Region;
	Region.Include(itemColumnRect);
	owner->ConstrainClippingRegion(&Region);
	owner->SetHighColor(Black);

	// icon column
	if(columnIndex==0) {
	
		itemColumnRect.left +=2.0;
		itemColumnRect.right=itemColumnRect.left + 15.0;
		itemColumnRect.top +=2.0;
		if(Height() > 20.0) {
			itemColumnRect.top +=ceil((Height()-20.0) / 2.0);
		}
		itemColumnRect.bottom=itemColumnRect.top + 15.0;
		owner->SetDrawingMode(B_OP_OVER);
		if(iFile->SmallIcon()) {
			owner->DrawBitmap(iFile->SmallIcon(), BRect(0.0, 0.0, 15.0, 15.0), itemColumnRect);
		}
		owner->SetDrawingMode(B_OP_COPY);
	
	} else {

		owner->SetLowColor(White);
		owner->SetHighColor(Black);

		char strg[400];
		strcpy(strg, "");
		
		uint32 align=B_ALIGN_LEFT;
		switch (columnIndex) {
		
			// name column
			case 1 :
				if(selected) {
					owner->SetLowColor(Black);
					owner->SetHighColor(White);
				}
				strcpy(strg, iFile->Name());
				break;
				
			// size column
			case 2 :
				align=B_ALIGN_RIGHT;
				iFile->SizeDesc(strg);
				break;
				
			// modified column
			case 3 :
				iFile->ModifiedTmDesc(strg);
				break;
				
			// created column
			case 4 :
				iFile->CreatedTmDesc(strg);
				break;
				
			// kind column
			case 5 :
				strcpy(strg, iFile->MimeDesc());
				break;
				
			// path column
			case 6 :
				strcpy(strg, iFile->PathDesc());
				break;	
		}
		
		BPoint point;
		if(align==B_ALIGN_LEFT)
		{
			point=BPoint(itemColumnRect.left + 2.0, 
							itemColumnRect.top + iTextOffset);
		}
		else
		if(align==B_ALIGN_RIGHT)
		{
			BFont font;
			iFileView->GetFont(&font);
			float stringWidth=font.StringWidth(strg);
			point=BPoint(itemColumnRect.right - stringWidth - 2.0,
							itemColumnRect.top + iTextOffset);
		}

		if(strg) owner->DrawString(strg, point);
	}
	
	owner->ConstrainClippingRegion(NULL);

}
		
// custom update
void SFileItem::Update(BView *owner, const BFont *font)
{
	CLVListItem::Update(owner, font);
	
	be_plain_font->GetHeight(&iFontAttr);
	
	float FontHeight=ceil(iFontAttr.ascent) + ceil(iFontAttr.descent);
	iTextOffset=ceil(iFontAttr.ascent) + ((Height()-FontHeight) / 2.0);
}

SFileView::SFileView(BRect bounds, char *name)
 : ColumnListView(	bounds, 
 					&iContainerView, 
 					name,
 					B_FOLLOW_ALL,				// resizing mode
					B_WILL_DRAW | 	
					B_FRAME_EVENTS |			// flags		 
					B_NAVIGABLE,
					B_MULTIPLE_SELECTION_LIST,	// view type
					false,						// hierarchical	*/
					true,						// horizontal scrollbars
					true,						// vertical scrollbars
					true,						// scroll-view corner
					B_FANCY_BORDER),			// border style
   iScanThreadId(0),
   iScanThreadActive(0),
   iScanThreadTerminate(0),
   iTreeItem(0), 
   iUseName(true),
   iUseSize(true),
   iUseModified(true),
   iUseCreated(false),
   iUseKind(false),
   iUsePath(false),
   iColIcon(0),
   iColName(0),
   iColSize(0),
   iColModified(0),
   iColCreated(0),
   iColKind(0),
   iColPath(0),
   iFileScan(false)
{
	// build columns
	BuildColumns();
	
	// create and set invoke message
	SetInvocationMessage(new BMessage(MSG_FILEVIEW_INVOKE));
		
	// create semaphores
	iScanThreadActive=create_sem(1, "FileView::ScanThreadActive");
	iScanThreadTerminate=create_sem(1, "FileView::ScanThreadTerminate");
	
	// set sorting function
	SetSortFunction(Compare);
	
	// set default sortkey (by name)
	SetSortKey(1);
}				

SFileView::~SFileView()
{
	// make sure we stop the scan thread before we go and delete ourselves
	RequestScanTerminate();

	if(iScanThreadActive) delete_sem(iScanThreadActive);
	if(iScanThreadTerminate) delete_sem(iScanThreadTerminate);	
}

// implemented MouseDown handler
void SFileView::MouseDown(BPoint point)
{
	BPoint tPoint;
	uint32 buttons;

	int32 itemIndex=IndexOf(point);

	// get mouse buttons
	GetMouse(&tPoint, &buttons, false);

	// right mouse button (only) ?
	if((itemIndex >=0) && (buttons & B_SECONDARY_MOUSE_BUTTON))
	{
	
		// grab focus
		MakeFocus();

		SFileItem *item=(SFileItem *) ItemAt(itemIndex);
	
		// select the item if it is part of a selected group
		if(!item->IsSelected())
			Select(itemIndex);

		bool folderOnly=false;
		
		SFileList fileList;
		CurrentSelections(&fileList);
		if(fileList.CountItems()==1)
		{
			SFile *file=(SFile *) fileList.ItemAt(0);
			folderOnly=file->IsDirectory();
		}

		bool hvTrash=false;
		
		for (int32 i=0; i < fileList.CountItems(); i++)
		{
			SFile *file=(SFile *) fileList.ItemAt(i);
			if(file->IsTrash())
			{
				hvTrash=true;
				break;
			}
		}		

		// let the main window handle the menus
		BMessage *message;
		STreeItem *currentFolder=(STreeItem *) iTreeItem;
		SFileItem *fileitem=(SFileItem*)ItemAt(CurrentSelection());
		
		if(currentFolder->File()->IsTrash() || hvTrash)
			message=new BMessage(MSG_PIONEER_TRASH_POPUP_MENU);			
		else if(fileitem->File()->TargetType()==TYPE_DIRECTORY)
			message=new BMessage(MSG_PIONEER_FOLDER_POPUP_MENU);
		else
			message=new BMessage(MSG_PIONEER_FILE_POPUP_MENU);
		message->AddPoint("point", ConvertToScreen(point));
		message->AddPointer("item", item);
		message->AddInt8("fileview",1);
		
		Window()->Looper()->PostMessage(message);
		delete message;

	}
	else
		ColumnListView::MouseDown(point);
}

void SFileView::KeyDown(const char *bytes,int32 numbytes)
{
	// Call the inherited version for everything that we don't specifically support
	ColumnListView::KeyDown(bytes,numbytes);
}

// build columns
void SFileView::BuildColumns() 
{

	// create icon column
	if(!iColIcon) {						  
	
		iColIcon=new CLVColumn(	NULL,
							  		20.0,
							  		CLV_LOCK_AT_BEGINNING |
							  		CLV_NOT_MOVABLE | 
							  		CLV_NOT_RESIZABLE | 
							  		CLV_PUSH_PASS | 
							  		CLV_MERGE_WITH_RIGHT);
		this->AddColumn(iColIcon);
	}
	
	// name column
	if(!iColName)
	{
		iColName=new CLVColumn("Name",110.0,CLV_SORT_KEYABLE,50.0);
		this->AddColumn(iColName);
	}
	iColName->SetShown(iUseName);
	
	// size column
	if(!iColSize)
	{
		iColSize=new CLVColumn("Size",80.0,CLV_SORT_KEYABLE,50.0);
		this->AddColumn(iColSize);
	
	}
	iColSize->SetShown(iUseSize);
		
	// modified column
	if(!iColModified)
	{
		iColModified=new CLVColumn("Modified",140.0,CLV_SORT_KEYABLE,50.0);
		this->AddColumn(iColModified);
	}
	iColModified->SetShown(iUseModified);

	// created column
	if(!iColCreated)
	{
		iColCreated=new CLVColumn("Created",140.0,CLV_SORT_KEYABLE,50.0);
		this->AddColumn(iColCreated);
	}
	iColCreated->SetShown(iUseCreated);
	
	// kind column
	if(!iColKind)
	{
		iColKind=new CLVColumn("Kind",80.0,CLV_SORT_KEYABLE,50.0);
		this->AddColumn(iColKind);
	}
	iColKind->SetShown(iUseKind);
		
	// path column
	if(!iColPath)
	{
		iColPath=new CLVColumn("Path",120.0,CLV_SORT_KEYABLE,50.0);
		this->AddColumn(iColPath);
	}
	iColPath->SetShown(iUsePath);
	
}

// set columns
void SFileView::SetColumns(bool useName,bool useSize,bool useModified,
	bool useCreated,bool useKind,bool usePath)
{
	iUseName=useName;
	iUseSize=useSize;
	iUseModified=useModified;
	iUseCreated=useCreated;
	iUseKind=useKind;
	iUsePath=usePath;

	BuildColumns();
}

// sorting function
int SFileView::Compare(const CLVListItem* aItem1, const CLVListItem* aItem2, int32 sortKey)
{
	SFileItem *item1=(SFileItem *) aItem1;
	SFileItem *item2=(SFileItem *) aItem2;
	char kind1[B_MIME_TYPE_LENGTH], kind2[B_MIME_TYPE_LENGTH];
	char path1[B_PATH_NAME_LENGTH], path2[B_PATH_NAME_LENGTH];

	// sort names
	switch (sortKey)
	{
	
		case 1 :	// sort on name
		{
			// Sort folders first
			if(item1->File()->TargetType()==TYPE_DIRECTORY && item2->File()->TargetType()==TYPE_DIRECTORY)
				// sort directories by name
				return strcasecmp(item1->File()->Name(), item2->File()->Name());

			if(item1->File()->TargetType()==TYPE_DIRECTORY)
				return -1;
			if(item2->File()->TargetType()==TYPE_DIRECTORY)
				return 1;

			return strcasecmp(item1->File()->Name(), item2->File()->Name());
		}			
		case 2 :	// sort on size
		{
			if(item1->File()->TargetType()==TYPE_DIRECTORY && item2->File()->TargetType()==TYPE_DIRECTORY)
				// sort directories by name
				return strcasecmp(item1->File()->Name(), item2->File()->Name());

			if(item1->File()->TargetType()==TYPE_DIRECTORY)
				return -1;
			if(item2->File()->TargetType()==TYPE_DIRECTORY)
				return 1;
			
				return item1->File()->Size() < item2->File()->Size() ? -1 : 1;				
		}
		case 3 :	// sort on modified date
		{
			return item1->File()->ModifiedTime() < item2->File()->ModifiedTime() ? -1 : 1;
		}			
		case 4 : 	// sort on created date
		{
			return item1->File()->CreatedTime() < item2->File()->CreatedTime() ? -1 : 1;
		}
		case 5 : 	// file type, sort folders first, subsort all alphabetically
		{
			if(item1->File()->TargetType()==TYPE_DIRECTORY && item2->File()->TargetType()==TYPE_DIRECTORY)
				// sort directories by name
				return strcasecmp(item1->File()->Name(), item2->File()->Name());

			if(item1->File()->TargetType()==TYPE_DIRECTORY)
				return -1;
			if(item2->File()->TargetType()==TYPE_DIRECTORY)
				return 1;

			strcpy(kind1, item1->File()->MimeDesc());
			strcpy(kind2, item2->File()->MimeDesc());
			int8 cmpval=strcasecmp(kind1, kind2);
			return (cmpval==0)?strcasecmp(item1->File()->Name(), item2->File()->Name()):cmpval;
		}
		case 6 :	// path
		{
			strcpy(path1, item1->File()->PathDesc());
			strcpy(path2, item2->File()->PathDesc());
			return strcasecmp(path1, path2);
		}
	}
	
	return 0;
}

// threaded file scan function
int32 SFileView::ScanThread(void *dataPtr)
{
	SFileView *view=(SFileView *) dataPtr;

	// acquire lock on 'active' semaphore
	acquire_sem(view->iScanThreadActive);

	// scan contents
	BDirectory directory;
	BEntry entry;

	// set directory to the entry represented by this item
	directory.SetTo(&view->iScanEntryRef);

	// scan entries
	bool terminated=false;
	sem_info semInfo;

	while ((directory.GetNextEntry(&entry)==B_OK)) {

		// owner thread requested we abort ?
		get_sem_info(view->iScanThreadTerminate, &semInfo);
		if(semInfo.count <=0) {
			terminated=true;
			break;
		}
			
		// if RequestScanTerminate is waiting for us to quit, it is
		// taking up a lock.  Hello deadlock.
		while (view->Window()->LockWithTimeout(10000)==B_TIMED_OUT) {
			get_sem_info(view->iScanThreadTerminate, &semInfo);
			if(semInfo.count <=0) {
				terminated=true;
				break;
			}
		}
		if(terminated) {
			break;
		}

		// add item to list	
		SFile file(&entry);
		if(!file.IsDesktop()) {
			SFileItem *item=new SFileItem(view, &file);
			view->AddItem(item);
		}
		view->Window()->Unlock();
	}

	directory.Unset();

	// try locking the window one more time so we can sort the new entries
	terminated=false;
	while (view->Window()->LockWithTimeout(10000)==B_TIMED_OUT) {
		get_sem_info(view->iScanThreadTerminate, &semInfo);
		if(semInfo.count <=0) {
			terminated=true;
			break;
		}
	}
	if(!terminated) {
		view->SortItems();
		view->Window()->Unlock();
	}

	// lower flag
	release_sem(view->iScanThreadActive);

	if(terminated) {
		return -1;
	} else {	
		return 0;
	}
}

// request scan termination and wait for it to stop
void SFileView::RequestScanTerminate()
{
	acquire_sem(iScanThreadTerminate);
	
	// sit and wait for the thread to end
	acquire_sem(iScanThreadActive);
	release_sem(iScanThreadActive);
	
	release_sem(iScanThreadTerminate);
}

// refresh tree
void SFileView::Refresh(BMessage *message)
{
	// if the scan thread is running, wait for it to stop
	RequestScanTerminate();

	// delete old items
	SFileItem *oldItem;
	for (uint32 i=0; (oldItem=(SFileItem *) ItemAt(i)); i++)
		delete oldItem;

	// clear list
	MakeEmpty();

	if(message==NULL)
		return;

	entry_ref entryRef;
	if(message->FindRef("root", &entryRef) !=B_OK)
		return;
	
	if(message->FindPointer("tree_item", (void **) &iTreeItem) !=B_OK)
		return;

	// update title on main window
	BMessage *titleMessage=new BMessage(MSG_PIONEER_UPDATE_TITLE);

	BEntry entry(&entryRef);
	BPath path(&entry);
	titleMessage->AddString("path", path.Path());
	Window()->Looper()->PostMessage(titleMessage);
	delete titleMessage;	 
		
	// start thread
	iScanEntryRef=	entryRef;
	iScanThreadId=spawn_thread(ScanThread, "FileView::ScanThread", 10, (void *) this);	
	resume_thread(iScanThreadId);
}

// current selections
void SFileView::CurrentSelections(SFileList *fileList)
{
	if(!fileList) {
		return;
	}

	for (int32 i=0; i < CountItems(); i++) {
		SFileItem *item=(SFileItem *) ItemAt(i); 
		if(item->IsSelected()) {
			fileList->AddItem((void *) item->File());
		}		
	}
}

// open file or directory
void SFileView::Open(BMessage *message)
{
	if(message==NULL) {
		return;
	}
	
	int32 itemIndex;
	if(message->FindInt32("index", &itemIndex) !=B_OK) {
		return;
	}
	
	SFileItem *item;
	item=(SFileItem *) ItemAt(itemIndex);
	if(item==NULL) {
		return;
	}
	
	// is a directory ?
	if(item->File()->IsDirectory()) {

		// sync with tree-view
		BMessage *message=new BMessage(MSG_TREEVIEW_EXPAND);
		message->AddPointer("tree_item", iTreeItem);
		message->AddRef("dir", item->File()->Ref());
		Looper()->PostMessage(message);
		delete message;

	} else {

		be_roster->Launch((entry_ref *) item->File()->Ref());
	}
}

// select all entries
void SFileView::SelectAll()
{
	// This is locking the window, and the scan thread won't continue until the lock is lifted.
	sem_info semInfo;
	get_sem_info(iScanThreadActive, &semInfo);
	if(semInfo.count <=0) {
		return;
	}

	for (int32 i=0; i < CountItems(); i++) {
		SFileItem *item=(SFileItem *) ItemAt(i); 
		if(!item->IsSelected()) {
			Select(i, true);
		}		
	}
}
		
// invert selection
void SFileView::InvertSelection()
{
	// This is locking the window, and the scan thread won't continue until the lock is lifted.
	sem_info semInfo;
	get_sem_info(iScanThreadActive, &semInfo);
	if(semInfo.count <=0) {
		return;
	}

	for (int32 i=0; i < CountItems(); i++) {
		SFileItem *item=(SFileItem *) ItemAt(i); 
		if(item->IsSelected()) {
			Deselect(i);
		} else {
			Select(i, true);
		}		
	}
}

// node monitor: scan for created
void SFileView::ScanForCreated(const SFile *file)
{
	entry_ref entryRef=*file->Ref();

	// find the parent of the normal entry
	BEntry childEntry(&entryRef);
	BEntry parentEntry;
	childEntry.GetParent(&parentEntry);
	
	entry_ref parentEntryRef;
	parentEntry.GetRef(&parentEntryRef);
	
	// the file activity applies to this list ?
	if(parentEntryRef !=iScanEntryRef) {
		return;
	}

	// check to see if we've already added it to the list
	// this prevents duplicate entries from being created
	// when we rename a folder
	SFileItem *existItem;
	bool found=false;
	for (uint32 i=0; (existItem=(SFileItem *) ItemAt(i)); i++) {
	
		// found it ?
		if(strcasecmp(existItem->File()->Name(), entryRef.name)==0) {
			found=true;
			break;
		}
	}

	if(found) {
		return;
	}
	
	// add the entry
	SFileItem *newItem=new SFileItem(this, file);																
	AddItem(newItem);
	
	SortItems();
}

// node monitor: scan for moved
void SFileView::ScanForMoved(const SFile *fromFile, const SFile *toFile)
{
	// use the standard creation routine
	ScanForCreated(toFile);

	// since we don't get a stupid parent node, we can't tell whether this
	// should be refreshed...STUPID STUPID STUPID
	ScanAllForRemoved();
}

// node monitor: scan for removed
void SFileView::ScanAllForRemoved()
{
	SFileItem *item;
	for (uint32 i=0; (item=(SFileItem *) ItemAt(i)); i++) {
	
		if(!item->File()->Exists()) {
		
			// remove it !
			RemoveItem(item);
			
			delete item;
		}	
	}
}

void SFileView::SpawnTracker()
{
	SFileItem *item=(SFileItem*)ItemAt(CurrentSelection());

	if(!item)
		return;

	BString string("/boot/beos/system/Tracker '");
	string+=item->File()->PathDesc();
	string+="' &";
	system(string.String());
	printf("%s\n",string.String());
}
