#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <fs_attr.h>

#include <Node.h>
#include <NodeInfo.h>
#include <Entry.h>
#include <Directory.h>
#include <Alert.h>

#include "Colors.h"

#include "SeekerWindow.h"
#include "SFileInfo.h"

#include "SFileWorker.h"

SFileWorker::SFileWorker(SeekerWindow *mainWindow)
 : iWindow(mainWindow),
   iFileOp(P_OP_NONE),
   iSrcFolder(0),
   iFileList(0)
{
	/* create new list */
	iFileList=new BList();
}

SFileWorker::~SFileWorker()
{
	if(iSrcFolder) delete iSrcFolder;

	if(iFileList) {
		for (int32 i=0; i < iFileList->CountItems(); i++) {
			entry_ref *entryRef=(entry_ref *) iFileList->ItemAt(i);
			delete entryRef;
		}
		iFileList->MakeEmpty();
		
		delete iFileList;
	}
}

void SFileWorker::NewFolder(const SFile *parentFolder)
{
	/* display modal dialog */
	if(parentFolder) {
		PNewFolderWindow *newFolderWindow=new PNewFolderWindow(iWindow, parentFolder);
		newFolderWindow->Show();
	}
}

void SFileWorker::GetInfo(const SFile *file)
{
	if(file) {

		SFileInfoWindow *infoWindow=new SFileInfoWindow(iWindow, file);
		iWindow->AddToSubset(infoWindow);
		infoWindow->Show();

	}
}

void SFileWorker::EditName(const SFile *file)
{
	if(file) {
	
		PRenameWindow *renameWindow=new PRenameWindow(iWindow, file);
		renameWindow->Show();
	}	
}

void SFileWorker::MoveToTrash(const SFile *srcFolder, const SFileList *fileList)
{
	BEntry entry(TRASH_ENTRY);
	SFile trashFolder(&entry);

	Cut(srcFolder, fileList);
	Paste(&trashFolder);
}

/* used for emptying trash (deleting in reverse recursive order),
   sorts by the directory depth */
static int CompareByLevel(const void *item1, const void *item2)
{
	const SFile *file1=(const SFile *) item1;
	const SFile *file2=(const SFile *) item2;
	
	return file1->Level() < file2->Level() ? -1 : 1;		
}

void SFileWorker::EmptyTrash(void)
{
	if(iSrcFolder)
	{
		delete iSrcFolder;
		iSrcFolder=0;
	}
	
	BEntry trashEntry(TRASH_ENTRY);
	iSrcFolder=new SFile(&trashEntry);

	for (int32 i=0; i < iFileList->CountItems(); i++) 
	{
		SFile *file=(SFile *) iFileList->ItemAt(i);
		delete file;
	}
	iFileList->MakeEmpty();

	// set directory to the entry represented by this item
	BEntry entry;
	BDirectory directory(TRASH_ENTRY);

	while ((directory.GetNextEntry(&entry, true)==B_OK))
	{
		SFile *file=new SFile(&entry);
		iFileList->AddItem(file);
	}

	// deep-scan
	int32 nItems=iFileList->CountItems();
	for (int32 indx=0; indx < nItems; indx++)
	{
		SFile *file=(SFile *) iFileList->ItemAt(indx);
		
		if(file->IsDirectory())
		{
			directory.SetTo(file->Ref());
			while (directory.GetNextEntry(&entry, true)==B_OK)
			{
				SFile *newFile=new SFile(&entry);				
				iFileList->AddItem(newFile);
				nItems +=1;				
			}
		}
	}

	iFileList->SortItems(CompareByLevel);

	iFileOp=P_OP_CUT;

	if(iFileList->CountItems()==0)
		return;

	PCopyWindow *copyWindow=new PCopyWindow(iWindow, iSrcFolder, iFileList, iSrcFolder, 
											   iFileOp, P_OP_REMOVE);
	iWindow->AddToSubset(copyWindow);
	copyWindow->Show();
	
	/* clear old list */
	for (int32 i=0; i < iFileList->CountItems(); i++) {
		SFile *file=(SFile *) iFileList->ItemAt(i);
		delete file;
	}
	iFileList->MakeEmpty();
}

/* add files to the copy list */
void SFileWorker::Copy(const SFile *srcFolder, const SFileList *fileList)
{
	/* clear old base-folder */
	if(iSrcFolder) {
		delete iSrcFolder;
		iSrcFolder=0;
	}
	
	iSrcFolder=new SFile(srcFolder);

	/* clear old list */
	for (int32 i=0; i < iFileList->CountItems(); i++) {
		SFile *file=(SFile *) iFileList->ItemAt(i);
		delete file;
	}
	iFileList->MakeEmpty();

	/* copy new list */	
	for (int32 j=0; j < fileList->CountItems(); j++) {
		SFile *file=new SFile((SFile *) fileList->ItemAt(j));
		iFileList->AddItem(file);
	}
	
	/* set operation */
	iFileOp=P_OP_COPY;
}

void SFileWorker::Cut(const SFile *srcFolder, const SFileList *fileList)
{
	/* same as copy but we'll override the file operation flag */
	Copy(srcFolder, fileList);
	iFileOp=P_OP_CUT;
}

/* start a file operation */
void SFileWorker::Paste(const SFile *folder)
{
	/* empty list ? */
	if(iFileList->CountItems()==0) {
		return;
	}

	PCopyWindow *copyWindow=new PCopyWindow(iWindow, iSrcFolder, iFileList, folder, 
											   iFileOp, P_OP_PASTE);
	iWindow->AddToSubset(copyWindow);
	copyWindow->Show();
	
	/* clear old list */
	for (int32 i=0; i < iFileList->CountItems(); i++) {
		SFile *file=(SFile *) iFileList->ItemAt(i);
		delete file;
	}
	iFileList->MakeEmpty();
}

void SFileWorker::PasteLink(const SFile *folder)
{
	/* empty list ? */
	if(iFileList->CountItems()==0) {
		return;
	}
	
	/* force copy operation to copy */
	iFileOp=P_OP_COPY;	

	PCopyWindow *copyWindow=new PCopyWindow(iWindow, iSrcFolder, iFileList, folder, 
											   iFileOp, P_OP_PASTE_LINK);
	iWindow->AddToSubset(copyWindow);
	copyWindow->Show();
	
	/* clear old list */
	for (int32 i=0; i < iFileList->CountItems(); i++) {
		SFile *file=(SFile *) iFileList->ItemAt(i);
		delete file;
	}
	iFileList->MakeEmpty();
}

PNewFolderWindow::PNewFolderWindow(SeekerWindow *mainWindow,
			 	 					const SFile *parentFile)
 : BWindow(BRect(50, 50, 0, 0), "New Folder", 
 			B_FLOATING_WINDOW,
			B_WILL_DRAW | B_NOT_ZOOMABLE | 
			B_NOT_MINIMIZABLE | B_NOT_RESIZABLE),
   iMainWindow(mainWindow),
   iView(0),
   iParentTextControl(0),
   iNameTextControl(0),
   iOkButton(0),
   iCancelButton(0)
{
	char *parentLabel="Parent";
	char *nameLabel="New Folder";

	iParentFile=new SFile(parentFile);

	/* set window feel */
	SetFeel(B_MODAL_APP_WINDOW_FEEL);

	/* create view */
	iView=new BView(Bounds(), "NewFolderWindow::View", B_FOLLOW_ALL, 
					   B_WILL_DRAW);
	iView->SetViewColor(BeBackgroundGrey);
	AddChild(iView);

	/* parent label/text */
	BRect rect;
	rect.left=10;
	rect.top=10;
	iParentTextControl=new BTextControl(rect, "NewFolderWindow::ParentTextControl",
										   parentLabel, iParentFile->Name(), NULL);
	iParentTextControl->ResizeToPreferred();
	iView->AddChild(iParentTextControl);

	/* new folder name label/text */
	rect.top +=10 + iParentTextControl->Bounds().Height();
	iNameTextControl=new BTextControl(rect, "NewFolderWindow::NameTextControl",
										 nameLabel, "", NULL);
	iNameTextControl->ResizeToPreferred();
	iView->AddChild(iNameTextControl);

	/* align fields */
	float parentStringWidth=be_plain_font->StringWidth(parentLabel);
	float nameStringWidth=be_plain_font->StringWidth(nameLabel);
	float maxWidth=MAX(parentStringWidth, nameStringWidth) + 10;
	
	iNameTextControl->SetDivider(maxWidth);
	iNameTextControl->ResizeTo(MAX(iNameTextControl->Bounds().Width(), 200),
								iNameTextControl->Bounds().Height());

	iParentTextControl->SetDivider(maxWidth);
	iParentTextControl->ResizeTo(iNameTextControl->Bounds().Width(), 
								  iParentTextControl->Bounds().Height()); 

	/* cancel button */
	iCancelButton=new BButton(BRect(), "NewFolderWindow::CancelButton",
								 "Cancel",  new BMessage(MSG_NEWFOLDER_WINDOW_CANCEL));
	iCancelButton->ResizeToPreferred();
	iCancelButton->MoveTo(iNameTextControl->Frame().right - iCancelButton->Bounds().Width(),
						   iNameTextControl->Frame().bottom + 10);			
				
	/* ok button */
	iOkButton=new BButton(BRect(), "NewFolderWindow::OkButton",
							 "OK",  new BMessage(MSG_NEWFOLDER_WINDOW_OK));
	iOkButton->ResizeToPreferred();
	iOkButton->MoveTo(iCancelButton->Frame().left - 10 - iOkButton->Bounds().Width(),
					   iCancelButton->Frame().top);	
	iOkButton->MakeDefault(true);
	
	/* add window in tab-order */
	iView->AddChild(iOkButton);
	iView->AddChild(iCancelButton);

	/* re-size window */
	rect.right=iCancelButton->Frame().right + 10;
	rect.bottom=iCancelButton->Frame().bottom + 10;
	ResizeTo(rect.right, rect.bottom);
	
	/* disable editing of parent name */
	iParentTextControl->TextView()->MakeEditable(false);
	
	/* set focus to name field */
	iNameTextControl->MakeFocus();
}			 	  

PNewFolderWindow::~PNewFolderWindow()
{
	if(iParentFile) delete iParentFile;
}

void PNewFolderWindow::MessageReceived(BMessage* message)
{
	switch(message->what)
	{
		case MSG_NEWFOLDER_WINDOW_OK		:
			CreateFolder();
			break;
	
		case MSG_NEWFOLDER_WINDOW_CANCEL	:
			Quit();
			break;

		default:
			BWindow::MessageReceived(message);
	}

}

void PNewFolderWindow::CreateFolder()
{
	char newPath[B_PATH_NAME_LENGTH];
		
	strcpy(newPath, iParentFile->PathDesc());
	strcat(newPath, "/");
	strcat(newPath, iNameTextControl->Text());

	BDirectory directory;
	status_t rc;
	
	rc=directory.CreateDirectory(newPath, 0);
	
	if(rc !=B_OK) {

		char errMsg[80];
		switch (rc) {
		
			case B_BAD_VALUE	 :
				strcpy(errMsg, "The folder cannot be created because an invalid folder name was specified (possibly because of illegal characters)");
				break;

			case B_FILE_EXISTS	 :
				strcpy(errMsg, "The folder cannot be created because the name specified is already used by another file or folder");
				break;
				
			case B_NAME_TOO_LONG :		
				strcpy(errMsg, "The folder cannot be created because the name specified is too long for the target file system");
				break;
				
			case B_NOT_ALLOWED	 :
				strcpy(errMsg, "The folder cannot be created because the target volume is read-only");
				break;
				
			case B_PERMISSION_DENIED :
				strcpy(errMsg, "The folder cannot be created because you do not have sufficient privileges on the target volume or parent folder");
				break;
				
			default				:
				sprintf(errMsg, "The folder cannot be created because an error occurred for which we do not have an explanation (RC=%ld)", rc);
		}
		
	
		BAlert *alert=new BAlert("Error", errMsg, "OK");
		alert->Go();
		return;
		
	} else {
		Quit();
	}
}		

PRenameWindow::PRenameWindow(SeekerWindow *mainWindow,
			 	 			  const SFile *file)
 : BWindow(BRect(50, 50, 0, 0), "Edit name", 
 			B_FLOATING_WINDOW,
			B_WILL_DRAW | B_NOT_ZOOMABLE | 
			B_NOT_MINIMIZABLE | B_NOT_RESIZABLE),
   iMainWindow(mainWindow),
   iView(0),
   iOldNameTextControl(0),
   iNewNameTextControl(0),
   iOkButton(0),
   iCancelButton(0)
{
	char *oldNameLabel="Old name";
	char *newNameLabel="New name";

	iFile=new SFile(file);

	/* set window feel */
	SetFeel(B_MODAL_APP_WINDOW_FEEL);

	/* create view */
	iView=new BView(Bounds(), "RenameWindow::View", B_FOLLOW_ALL, 
					   B_WILL_DRAW);
	iView->SetViewColor(BeBackgroundGrey);
	AddChild(iView);

	/* old-name label/text */
	BRect rect;
	rect.left=10;
	rect.top=10;
	iOldNameTextControl=new BTextControl(rect, "RenameWindow::OldNameTextControl",
										    oldNameLabel, iFile->Name(), NULL);
	iOldNameTextControl->ResizeToPreferred();
	iView->AddChild(iOldNameTextControl);

	/* new folder name label/text */
	rect.top +=10 + iOldNameTextControl->Bounds().Height();
	iNewNameTextControl=new BTextControl(rect, "RenameWindow::NewNameTextControl",
										 	newNameLabel, "", NULL);
	iNewNameTextControl->ResizeToPreferred();
	iView->AddChild(iNewNameTextControl);

	/* align fields */
	float oldNameStringWidth=be_plain_font->StringWidth(oldNameLabel);
	float newNameStringWidth=be_plain_font->StringWidth(newNameLabel);
	float maxWidth=MAX(oldNameStringWidth, newNameStringWidth) + 10;
	
	iNewNameTextControl->SetDivider(maxWidth);
	iNewNameTextControl->ResizeTo(MAX(iNewNameTextControl->Bounds().Width(), 200),
								   iNewNameTextControl->Bounds().Height());

	iOldNameTextControl->SetDivider(maxWidth);
	iOldNameTextControl->ResizeTo(iNewNameTextControl->Bounds().Width(), 
								   iOldNameTextControl->Bounds().Height()); 

	/* cancel button */
	iCancelButton=new BButton(BRect(), "RenameWindow::CancelButton",
								 "Cancel",  new BMessage(MSG_RENAME_WINDOW_CANCEL));
	iCancelButton->ResizeToPreferred();
	iCancelButton->MoveTo(iNewNameTextControl->Frame().right - iCancelButton->Bounds().Width(),
						   iNewNameTextControl->Frame().bottom + 10);			
				
	/* ok button */
	iOkButton=new BButton(BRect(), "RenameWindow::OkButton",
							 "OK",  new BMessage(MSG_RENAME_WINDOW_OK));
	iOkButton->ResizeToPreferred();
	iOkButton->MoveTo(iCancelButton->Frame().left - 10 - iOkButton->Bounds().Width(),
					   iCancelButton->Frame().top);	
	iOkButton->MakeDefault(true);
	
	/* add window in tab-order */
	iView->AddChild(iOkButton);
	iView->AddChild(iCancelButton);

	/* re-size window */
	rect.right=iCancelButton->Frame().right + 10;
	rect.bottom=iCancelButton->Frame().bottom + 10;
	ResizeTo(rect.right, rect.bottom);
	
	/* disable editing of old-name */
	iOldNameTextControl->TextView()->MakeEditable(false);
	iOldNameTextControl->TextView()->SetViewColor(ui_color(B_PANEL_BACKGROUND_COLOR));
	
	/* set focus to name field */
	iNewNameTextControl->MakeFocus();
}			 	  

PRenameWindow::~PRenameWindow()
{
	if(iFile) delete iFile;
}

void PRenameWindow::MessageReceived(BMessage* message)
{
	switch(message->what)
	{
		case MSG_RENAME_WINDOW_OK		:
			RenameFile();
			break;
	
		case MSG_RENAME_WINDOW_CANCEL	:
			Quit();
			break;

		default:
			BWindow::MessageReceived(message);
	}

}

void PRenameWindow::RenameFile()
{
	BEntry entry(iFile->Ref());
	
	status_t rc;
	
	rc=entry.Rename(iNewNameTextControl->Text(), false);
	
	if(rc !=B_OK) {

		char errMsg[80];
		switch (rc) {
		
			case B_FILE_EXISTS	 :
				strcpy(errMsg, "The file or folder cannot be renamed because the name specified is already used by another file or folder");
				break;
				
			default				:
				sprintf(errMsg, "The folder cannot be created because an error occurred for which we do not have an explanation (RC=%ld)", rc);
		}
		
	
		BAlert *alert=new BAlert("Error", errMsg, "OK");
		alert->Go();
		return;
		
	} else {
		Quit();
	}
}

PCopyWindow::PCopyWindow(SeekerWindow *mainWindow, 
						  SFile *srcFolder, BList *fileList, 
						  const SFile *destFolder, int copyOp, int pasteOp)
 : BWindow(windowFrame(), "Progress", B_FLOATING_WINDOW,B_WILL_DRAW),
   iCopyThreadId(0),
   iCopyThreadActive(0),
   iCopyThreadTerminate(0),
   iMainWindow(mainWindow),
   iView(0),
   iStatusBar(0),
   iCancelButton(0),
   iSrcFolder(0),
   iFileList(0),
   iDestFolder(0),
   iCopyOp(copyOp),
   iPasteOp(pasteOp)
{
	iCopyThreadActive=create_sem(1, "CopyWindow::CopyThreadActive");
	iCopyThreadTerminate=create_sem(1, "CopyWindow::CopyThreadTerminate");

	iSrcFolder=new SFile(srcFolder);

	iFileList=new BList(fileList->CountItems());
	for (int32 i=0; i < fileList->CountItems(); i++)
	{
		SFile *file=new SFile((SFile *) fileList->ItemAt(i));
		iFileList->AddItem(file);
	}
	
	iDestFolder=new SFile(destFolder);

	iView=new BView(Bounds(), "CopyWindow::View", B_FOLLOW_ALL, B_WILL_DRAW);
	iView->SetViewColor(BeBackgroundGrey);
	AddChild(iView);

	BRect statusBarBounds(10, 10, Bounds().right-10, 35);
	iStatusBar=new BStatusBar(statusBarBounds, "CopyWindow::StatusBar");
	iView->AddChild(iStatusBar);

	iStatusBar->SetMaxValue(iFileList->CountItems());

	BRect buttonBounds(211, 10, 0, 0);	
	iCancelButton=new BButton(buttonBounds, "CopyWindow::CancelButton",
			"Cancel",  new BMessage(MSG_COPY_WINDOW_CANCEL));
	iCancelButton->ResizeToPreferred();
	buttonBounds=iCancelButton->Frame();
	iCancelButton->MoveTo(Bounds().right-(buttonBounds.Width()+10),
		iStatusBar->Frame().bottom+10);
	iView->AddChild(iCancelButton);
	
	ResizeTo(iCancelButton->Frame().right+10,iCancelButton->Frame().bottom+10);
}

PCopyWindow::~PCopyWindow()
{
	/* make sure we stop the copy thread before we go and delete ourselves */
	RequestCopyTerminate();

	/* remove from subset */
	BMessage *message=new BMessage(MSG_PIONEER_REMOVE_FROM_SUBSET);
	message->AddPointer("window", this);
	iMainWindow->Looper()->PostMessage(message);
	delete message;

	if(iCopyThreadActive) delete_sem(iCopyThreadActive);
	if(iCopyThreadTerminate) delete_sem(iCopyThreadTerminate);	

	if(iSrcFolder) delete iSrcFolder;
	
	if(iFileList) {
		for (int32 i=0; i < iFileList->CountItems(); i++) {
			SFile *file=(SFile *) iFileList->ItemAt(i);
			delete file;
		}	
		iFileList->MakeEmpty();
		
		delete iFileList;
	}
	
	delete iDestFolder;
}

void PCopyWindow::MessageReceived(BMessage* message)
{
	switch(message->what)
	{
		case MSG_COPY_WINDOW_CANCEL		:
			RequestCopyTerminate();
			Quit();
			break;

		default:
			BWindow::MessageReceived(message);
	}
}

void PCopyWindow::Show(void)
{
	BWindow::Show();

	/* start the copy thread */
	iCopyThreadId=spawn_thread(CopyThread, "CopyWindow::CopyThread", 10, (void *) this);	
	resume_thread(iCopyThreadId);
}

/* threaded file scan function */
int32 PCopyWindow::CopyThread(void *dataPtr)
{
	PCopyWindow *window=(PCopyWindow *) dataPtr;

	/* acquire lock on 'active' semaphore */
	acquire_sem(window->iCopyThreadActive);

	sem_info semInfo;
	bool terminated=false;

	/* get item count */
	int32 nItems=0;
	while (window->LockWithTimeout(10000 /* 0.01s */)==B_TIMED_OUT) {
		get_sem_info(window->iCopyThreadTerminate, &semInfo);
		if(semInfo.count <=0) {
			terminated=true;
			break;
		}
	}
	if(!terminated) {
		nItems=window->iFileList->CountItems();
		window->Unlock();	
	} else {
		release_sem(window->iCopyThreadActive);
		return -1;
	}
	
	char text[B_FILE_NAME_LENGTH + 20], trailingText[80];

	bool cancelled=false;
	
	/* allows copying of the correct hierachery when auto-renaming of folders moved
	   to the trash or copied to the same folder */ 
	char autoRenameOld[B_PATH_NAME_LENGTH], autoRenameNew[B_PATH_NAME_LENGTH];
	strcpy(autoRenameOld, "");
	strcpy(autoRenameNew, "");
	
	for (int32 sourceIndx=0; (sourceIndx < nItems) && !cancelled; sourceIndx++) {
		
		/* if RequestCopyTerminate is waiting for us to quit, it is
		   taking up a lock.  Hello deadlock. */
		while (window->LockWithTimeout(10000 /* 0.01s */)==B_TIMED_OUT) {
			get_sem_info(window->iCopyThreadTerminate, &semInfo);
			if(semInfo.count <=0) {
				terminated=true;
				break;
			}
		}
		if(terminated) {
			break;
		}

		SFile *file=(SFile *) window->iFileList->ItemAt(sourceIndx);

		if(window->iCopyOp==P_OP_CUT) {
		
			if(window->iPasteOp==P_OP_PASTE) {
		
				BEntry srcEntry(file->Ref());
				BDirectory destDir(window->iDestFolder->Ref()); 
				BEntry destDirEntry(window->iDestFolder->Ref());
				BPath destPath(&destDirEntry);

				/* the following is code to allow for sub-structures to be copied with
				   their original hierarchy */
				char destPathName[B_PATH_NAME_LENGTH];
				strcpy(destPathName, destPath.Path());		
				{
					char filePath[B_PATH_NAME_LENGTH], srcPath[B_PATH_NAME_LENGTH];

					strcpy(filePath, file->PathDesc());
					strcpy(srcPath, window->iSrcFolder->PathDesc());
					
					uint32 srcPrfxLgth=strlen(srcPath);
					
					/* swap the first part of the tree if we've automagically updated it */
					uint32 autoRenameLgth=strlen(autoRenameOld);
					if(autoRenameLgth) {					
						strcat(destPathName, "/");
						strcat(destPathName, autoRenameNew);
						srcPrfxLgth +=autoRenameLgth + 1;
					}					
					
					strcat(destPathName, &filePath[ srcPrfxLgth ]);
 				}	

				bool clobber=false;
				do {

	  				/* test if destination exists first */
					BEntry destEntry(destPathName);
					
					if(destEntry.Exists()) {
	
						/* automagically change the name if we're pasting to the same directory
						   or moving to trash */
						if((strcasecmp(window->iDestFolder->PathDesc(), window->iSrcFolder->PathDesc())==0) || 
						     window->iDestFolder->IsTrash()) { 

							SFile::LeafFromPath(destPathName, autoRenameOld);
			
							char newDestPathName[B_PATH_NAME_LENGTH];	
							int32 attempt=0;
							do {
							
								if(attempt > 0) {
									sprintf(newDestPathName, "%s copy %ld", destPathName, attempt);								
								} else {
									sprintf(newDestPathName, "%s copy", destPathName);								
								}								
								destEntry.SetTo(newDestPathName);
								attempt +=1;
							
							} while (destEntry.Exists());
							
							strcpy(destPathName, newDestPathName);
							
							SFile::LeafFromPath(destPathName, autoRenameNew);
												
						} 					
					}
							
					status_t rc;
					rc=srcEntry.MoveTo(&destDir, destPathName, clobber);
					clobber=false;					

					/* handle errors */
					if(rc !=B_OK) {
	
						char errMsg[256];
						BAlert *alert;
						
						switch (rc) {
						
							case B_ENTRY_NOT_FOUND		:
								sprintf( errMsg, "The file or folder %s was moved or deleted  \
												   during this operation.  The operation will \
												   continue.", file->Name()); 
								alert=new BAlert("Error", errMsg, "OK", NULL, NULL, 
															B_WIDTH_AS_USUAL, B_WARNING_ALERT);
								alert->Go();
								break;
								
							case B_FILE_EXISTS			:
								sprintf( errMsg, "An item named \"%s\" already exists in this \
												   folder. Would you like to replace it with the \
												   file you are moving?", file->Name()); 
								alert=new BAlert("Error", errMsg, "Replace", "Cancel", "Cancel all");
								rc=alert->Go();
								if(rc==0) {
									clobber=true;
								} else if(rc==2) {
									cancelled=true;
								}
								break;							
								
							default						:
								sprintf(errMsg, "The file or folder %s was not moved because \
												  an error occurred for which we do not have \
												  an explanation (RC=%ld)", file->Name(), rc);
								alert=new BAlert("Error", errMsg, "Continue");
								alert->Go();
						}

					} 
		
				} while (clobber);
					
			} else if(window->iPasteOp==P_OP_REMOVE) {

				/* used for emptying the trash */			
				BEntry srcEntry(file->Ref());
				srcEntry.Remove();
				
			}
			
		} else if(window->iCopyOp==P_OP_COPY) {
		
			if(window->iPasteOp==P_OP_PASTE) {
			
				BDirectory destDir(window->iDestFolder->Ref());
				BEntry destDirEntry(window->iDestFolder->Ref());
				BPath destPath(&destDirEntry);

				/* the following is code to allow for sub-structures to be copied with
				   their original hierarchy */
				char destPathName[B_PATH_NAME_LENGTH];
				strcpy(destPathName, destPath.Path());		
				{
					char filePath[B_PATH_NAME_LENGTH], srcPath[B_PATH_NAME_LENGTH];

					strcpy(filePath, file->PathDesc());
					strcpy(srcPath, window->iSrcFolder->PathDesc());
					
					uint32 srcPrfxLgth=strlen(srcPath);
					
					/* swap the first part of the tree if we've automagically updated it */
					uint32 autoRenameLgth=strlen(autoRenameOld);
					if(autoRenameLgth) {
						strcat(destPathName, "/");
						strcat(destPathName, autoRenameNew);
						srcPrfxLgth +=autoRenameLgth + 1;
					}					
					
					strcat(destPathName, &filePath[ srcPrfxLgth ]);
 				}	

  				/* test if destination exists first */
				BEntry destEntry(destPathName);
				
				if(destEntry.Exists()) {

					/* automagically change the name if we're pasting to the same directory
					   or moving to trash */
					if((strcasecmp(window->iDestFolder->PathDesc(), window->iSrcFolder->PathDesc())==0) || 
					     window->iDestFolder->IsTrash()) { 

						SFile::LeafFromPath(destPathName, autoRenameOld);
						
						char newDestPathName[B_PATH_NAME_LENGTH];	
						int32 attempt=0;
						do {

							if(attempt > 0) {
								sprintf(newDestPathName, "%s copy %ld", destPathName, attempt);								
							} else {
								sprintf(newDestPathName, "%s copy", destPathName);								
							}								
							destEntry.SetTo(newDestPathName);
						
							attempt +=1;
						
						} while (destEntry.Exists());
	
						strcpy(destPathName, newDestPathName);
						
						SFile::LeafFromPath(destPathName, autoRenameNew);
										
					} else {
					
						char errMsg[256];
						BAlert *alert;
	
						sprintf( errMsg, "An item named \"%s\" already exists in this \
										   folder. Would you like to replace it with the \
										   file you are copying?", file->Name()); 
	
						alert=new BAlert("Error", errMsg, "Replace", "Cancel", "Cancel all");
						status_t rc=alert->Go();
						if(rc==0) {
							/* delete existing file */
							BEntry delEntry(destPathName);
							delEntry.Remove();
						} else if(rc==2) {
							cancelled=true;
						}
					
					}
				
				}
				
				if(file->IsDirectory()) {
				
					/* add children to bottom of list */
					BDirectory directory(file->Ref());
					BEntry entry;
					
					while (directory.GetNextEntry(&entry, true)==B_OK) {
						SFile *file=new SFile(&entry);				
						window->iFileList->AddItem(file);
						nItems +=1;						
					}

					/* create new directory instead of copying */
					destDir.CreateDirectory(destPathName, 0);

				} else {
				
					/* copy file */
					BFile srcFile(file->Ref(), B_READ_ONLY);
					BFile destFile(destPathName, B_WRITE_ONLY | B_CREATE_FILE);
					
					char buf[4096];
					ssize_t bytesRead=0;
					
					do {
	
						bytesRead=srcFile.Read((void *) buf, sizeof(buf));
						destFile.Write(buf, bytesRead);
					
					} while (bytesRead > 0);

					/* copy attributes */
					BNode srcNode(srcFile);
					BNode destNode(destFile);
					
					char attrName[B_ATTR_NAME_LENGTH];
					srcNode.RewindAttrs();
					while (srcNode.GetNextAttrName(attrName)==B_OK) {
					
						/* first get the size & type of the attribute */
						attr_info attrInfo;
						if(srcNode.GetAttrInfo(attrName, &attrInfo) !=B_OK) {
							continue;
						}
						
						/* allocate memory to hold the attribute */
						void *attrBuf=malloc(attrInfo.size); 
						if(attrBuf==0) {
							continue;
						}
						
						/* WE NEED SOME GOOD ERROR HANDLING CODE HERE */ 						
						ssize_t attrSz=0;
						attrSz=srcNode.ReadAttr(attrName, attrInfo.type, 0LL, attrBuf, attrInfo.size);
						
						destNode.WriteAttr(attrName, attrInfo.type, 0LL, attrBuf, attrSz);
						destNode.Sync();
																
						free(attrBuf);
					}

				}
			
			} else if(window->iPasteOp==P_OP_PASTE_LINK) {
		
				BEntry srcEntry(file->Ref());
				BPath srcPath(&srcEntry);
				const char *srcPathName=srcPath.Path();
				char srcLeafName[B_FILE_NAME_LENGTH];
				srcEntry.GetName(srcLeafName);

				BDirectory destDir(window->iDestFolder->Ref()); 

				BPath destPath(&destDir, srcPath.Leaf());
				const char *destPathName=destPath.Path();
								
				bool clobber=false;
				do {
				
					if(clobber) {
						/* delete existing file */
						BEntry delEntry(destPathName);
						delEntry.Remove();
						clobber=false;					
					}
							
					status_t rc;
					rc=destDir.CreateSymLink(destPathName, srcPathName, NULL);
			
					/* handle errors */
					if(rc !=B_OK) {
	
						char errMsg[256];
						BAlert *alert;
						
						switch (rc) {
						
							case B_ENTRY_NOT_FOUND		:
								sprintf( errMsg, "The file or folder %s was moved or deleted  \
												   during this operation.  The operation will \
												   continue.", file->Name()); 
								alert=new BAlert("Error", errMsg, "OK", NULL, NULL, 
															B_WIDTH_AS_USUAL, B_WARNING_ALERT);
								alert->Go();
								break;
								
							case B_FILE_EXISTS			:
								sprintf( errMsg, "An item named \"%s\" already exists in this \
												   folder. Would you like to replace it with the \
												   symbolic link you are creating?", file->Name()); 
								alert=new BAlert("Error", errMsg, "Replace", "Cancel", "Cancel all");
								rc=alert->Go();
								if(rc==0) {
									clobber=true;
								} else if(rc==2) {
									cancelled=true;
								}
								break;							
								
							default						:
								sprintf(errMsg, "The file or folder %s was not copied because \
												  an error occurred for which we do not have \
												  an explanation (RC=%ld)", file->Name(), rc);
								alert=new BAlert("Error", errMsg, "Continue");
								alert->Go();
						}

					} 

				} while (clobber);
				
			}			
		}

		/* update status bar */
		sprintf(text, "Copying: %s", file->Name());
		sprintf(trailingText, "%ld items remaining", nItems - sourceIndx - 1);
		window->iStatusBar->Update(1, text, trailingText);

		window->Unlock();
	
		snooze(100000);
	}

	/* close window */
	while (window->LockWithTimeout(10000 /* 0.01s */)==B_TIMED_OUT) {
		get_sem_info(window->iCopyThreadTerminate, &semInfo);
		if(semInfo.count <=0) {
			terminated=true;
			break;
		}
	}
	if(!terminated) {
		window->Looper()->PostMessage(B_QUIT_REQUESTED);
		window->Unlock();
	}

	release_sem(window->iCopyThreadActive);

	return (terminated)?-1:0;
}

/* request copy termination and wait for it to stop */
void PCopyWindow::RequestCopyTerminate()
{
	acquire_sem(iCopyThreadTerminate);
		
	/* sit and wait for the thread to end */
	acquire_sem(iCopyThreadActive);
	release_sem(iCopyThreadActive);
	
	release_sem(iCopyThreadTerminate);
}

