#include <alloc.h>
#include <stdio.h>
#include <string.h>
#include <Application.h>
#include <Window.h>
#include <Region.h>
#include <Node.h>
#include <NodeInfo.h>
#include <Directory.h>
#include <OutlineListView.h>
#include <String.h>
#include <ScrollView.h>
#include <ScrollBar.h>

#include <Alert.h>

#include "Colors.h"
#include "CLVColumn.h"

#include "SeekerWindow.h"
#include "STreeView.h"

STreeItem::STreeItem(STreeView *treeView, const SFile *file, bool expand)
 : CLVListItem(0, true, expand, 20.0),
   iScanThreadActive(0),
   iScanThreadTerminate(0),
   iTreeView(treeView),
   iFile(0),
   iTextOffset(0),
   iTreeScan(false),
   iFileScan(false)   		   
{
	iFile=new SFile(file);

	// create semaphores
	iScanThreadActive=create_sem(1, "TreeItem::ScanThreadActive");
	iScanThreadTerminate=create_sem(1, "TreeItem::ScanThreadTerminate");

	strcpy(iAutoSelectDir, "");
	
	// set level
	// WE SHOULD BE ABLE TO DO THIS MORE EFFICIENTLY	
	int32 level;
	if(!iFile->IsVolume()) {
		level=-3;
		BEntry tmpEntry(iFile->Ref());
		while (tmpEntry.GetParent(&tmpEntry)==B_OK) {
			// count levels by recursing tree
			level +=1;
		}
	} else { 
		// always under desktop
		level=1;
	}
	SetOutlineLevel(level);
	
	// add to node monitor watch list
	BNode node(iFile->Ref());
	node_ref nodeRef;
	node.GetNodeRef(&nodeRef);
	
	BMessage *message=new BMessage(MSG_NODEMONITOR_ADD_WATCH);
	message->AddInt32("device", nodeRef.device);
	message->AddInt64("node", nodeRef.node);
	iTreeView->Looper()->PostMessage(message);
	delete message;

}

STreeItem::~STreeItem()
{
	// remove from watch list
	BNode node(iFile->Ref());
	node_ref nodeRef;
	node.GetNodeRef(&nodeRef);

	BMessage *message=new BMessage(MSG_NODEMONITOR_REMOVE_WATCH);
	message->AddInt32("device", nodeRef.device);
	message->AddInt64("node", nodeRef.node);
	iTreeView->Looper()->PostMessage(message);
	delete message;

	// make sure we stop the scan thread before we go and delete ourselves
	RequestScanTerminate();

	if(iScanThreadActive) delete_sem(iScanThreadActive);
	if(iScanThreadTerminate) delete_sem(iScanThreadTerminate);	

	if(iFile) delete iFile;
}

// custom paint
void STreeItem::DrawItemColumn(BView *owner, BRect itemColumnRect, 
								int32 columnIndex, bool complete)
{
	rgb_color color;
	bool selected=IsSelected();
	
	if(selected) {
		color=Black;
	} else {
		color=White;
	}
	
	owner->SetLowColor(color);
	owner->SetDrawingMode(B_OP_COPY);
	if((selected || complete) && (columnIndex==2)) {
		owner->SetHighColor(color);

		BRect rect=itemColumnRect;
		rect.left -=1.0;
		rect.top +=iTextOffset - ceil(iFontAttr.ascent);
		rect.bottom -=((iTextOffset - ceil(iFontAttr.ascent)) / 2.0) + 2.0;
		rect.right=rect.left + owner->StringWidth(iFile->Name()) + 3.0;

		owner->FillRect(rect);
	} else if(complete && (columnIndex==0)) {
		// redraw expansion arrow
		owner->SetHighColor(White);
		owner->FillRect(itemColumnRect);
	}
	
	BRegion Region;
	Region.Include(itemColumnRect);
	owner->ConstrainClippingRegion(&Region);
	owner->SetHighColor(Black);

	// icon column
	if(columnIndex==1) {
	
		itemColumnRect.left +=2.0;
		itemColumnRect.right=itemColumnRect.left + 15.0;
		itemColumnRect.top +=2.0;
		if(Height() > 20.0) {
			itemColumnRect.top +=ceil((Height()-20.0) / 2.0);
		}
		itemColumnRect.bottom=itemColumnRect.top + 15.0;
		owner->SetDrawingMode(B_OP_OVER);
		if(iFile->SmallIcon()) {
			owner->DrawBitmap(iFile->SmallIcon(), BRect(0.0, 0.0, 15.0, 15.0), itemColumnRect);
		}
		owner->SetDrawingMode(B_OP_COPY);
	
	// label column
	} else if(columnIndex==2) {
		BPoint point=BPoint(itemColumnRect.left + 2.0, 
							   itemColumnRect.top + iTextOffset	);

		if(selected) {
			owner->SetHighColor(White);
		}

		owner->DrawString(iFile->Name(), point);
	}
	
	owner->ConstrainClippingRegion(NULL);
	
	// use this hook function to capture changes in expansion
	if(IsExpanded() && !iTreeScan) {

		iTreeScan=true;

		// if the scan thread is running, wait for it to stop 
		// (BTW, this shouldn't happen since iTreeScan should guarantee it is only run once)
		RequestScanTerminate();

		// start thread
		thread_id threadId=spawn_thread(ScanThread, "TreeItem::ScanThread", 10, (void *) this);	
		resume_thread(threadId);
	}

	// update file view
	if(selected && !iFileScan) {
		iFileScan=true;
		BMessage *message=new BMessage(MSG_FILEVIEW_REFRESH);
		message->AddPointer("tree_item", (void *) this);
		message->AddRef("root", iFile->Ref());
		iTreeView->Looper()->PostMessage(message);
		delete message;
	
	} else if(!selected) {
		iFileScan=false;
	}
}
		
// custom update
void STreeItem::Update(BView *owner, const BFont *font)
{
	CLVListItem::Update(owner, font);
	
	be_plain_font->GetHeight(&iFontAttr);
	
	float FontHeight=ceil(iFontAttr.ascent) + ceil(iFontAttr.descent);
	iTextOffset=ceil(iFontAttr.ascent) + ((Height()-FontHeight) / 2.0);
}

// update icon displayed (specifically for 'Trash')
void STreeItem::ResetIcon()
{
	iFile->ResetIcons();
	iTreeView->Invalidate();
}

// does the item have an active scan thread ?
bool STreeItem::IsScanning()
{
	sem_info semInfo;
	
	get_sem_info(iScanThreadActive, &semInfo);
	if(semInfo.count <=0) {
		return true;
	} else {
		return false;
	}
}	
		
// has the item been previously expanded ?
bool STreeItem::IsScanned()
{
	return iTreeScan;
}

// auto-select when scanning
void STreeItem::SetAutoSelectOnScanDir(const char *name)
{
	strcpy(iAutoSelectDir, name);
}

// expand the item and select a specific directory
void STreeItem::ExpandAndSelect(SFile *file)
{
	// wait for the scanning thread to stop
	acquire_sem(iScanThreadActive);
	release_sem(iScanThreadActive);	

	// has been expanded before ?
	bool wasScanned=IsScanned();
	if(!wasScanned) {
		// instruct thread to select the entry when it adds it to the list
		SetAutoSelectOnScanDir(file->Name());
	}
	
	// recursive expand
	BList *expandList=new BList();
	STreeItem *expandItem=this;
	
	while (true) {
		expandItem=(STreeItem *) iTreeView->Superitem(expandItem);
		if(expandItem==0) {
			// made it to the root
			break;
		}
		
		if(!expandItem->IsExpanded()) {
			expandList->AddItem(expandItem);	
		}
	}

	// go backwards through the list, expandin' all the way
	for (int32 i=expandList->CountItems(); i > 0; i--) {
		expandItem=(STreeItem *) expandList->ItemAt(i - 1);	
		iTreeView->Expand(expandItem);
	}
	
	delete expandList;
	
	// finally, expand ourself
	iTreeView->Expand(this);
	
	// the item should already exist in the list
	if(wasScanned) {

		// scan all subitems
		for (int32 i=0; i < iTreeView->FullListNumberOfSubitems(this); i++) {
			int32 itemIndex=iTreeView->FullListIndexOf(this) + i + 1;
			STreeItem *subItem=(STreeItem *) iTreeView->FullListItemAt(itemIndex);
			
			const entry_ref *subItemEntryRef=subItem->iFile->Ref();
	
			BNode primaryNode(file->Ref());
			BNode subItemNode(subItemEntryRef);

			if(primaryNode==subItemNode) {
				int32 visualIndex=iTreeView->IndexOf(subItem);
				iTreeView->Select(visualIndex);
				break;
			}
		}
	}
}

// threaded tree scan function
int32 STreeItem::ScanThread(void *dataPtr)
{
	STreeItem *item=(STreeItem *) dataPtr;

	// acquire lock on 'active' semaphore
	acquire_sem(item->iScanThreadActive);

	// if we didn't complete the scan the last time, remove the items
	sem_info semInfo;
	bool terminated=false;

	while (item->iTreeView->Window()->LockWithTimeout(10000)==B_TIMED_OUT) {
		get_sem_info(item->iScanThreadTerminate, &semInfo);
		if(semInfo.count <=0) {
			terminated=true;
			break;
		}
	}
	if(!terminated) {
		
		// for some reason or another, if we delete old items from the 'Desktop'
		// item then the first volume will not show up...
		if(item->OutlineLevel() > 0) {
	
			for (int32 i=0; i < item->iTreeView->FullListNumberOfSubitems(item); i++) {
				int32 itemIndex=item->iTreeView->FullListIndexOf(item) + i + 1;
				STreeItem *subItem=(STreeItem *) item->iTreeView->FullListItemAt(itemIndex);
			
				item->iTreeView->RemoveItem(subItem);
				delete subItem;
			}
		}
		item->iTreeView->Window()->Unlock();

	} else {
		// lower flag
		release_sem(item->iScanThreadActive);
	
		return -1;
	}
	
	BDirectory directory;
	BEntry entry;

	// set directory to the entry represented by this item
	directory.SetTo(item->iFile->Ref());

	// scan entries
	while (directory.GetNextEntry(&entry, true)==B_OK) {

		// owner thread requested we abort ?
		get_sem_info(item->iScanThreadTerminate, &semInfo);
		if(semInfo.count <=0) {
			terminated=true;
			break;
		}

		// skip non-directories
		if(!entry.IsDirectory()) {
			continue;
		}

		// if RequestScanTerminate is waiting for us to quit, it is
		// taking up a lock.  Hello deadlock.
		while (item->iTreeView->Window()->LockWithTimeout(10000)==B_TIMED_OUT) {
			get_sem_info(item->iScanThreadTerminate, &semInfo);
			if(semInfo.count <=0) {
				terminated=true;
				break;
			}
		}
		if(terminated) {
			break;
		}

		// add item to tree under current directory
		SFile file(&entry);
		if(file.IsDesktop()) {
			item->iTreeView->Window()->Unlock();
			continue;
		}
		
		STreeItem *subItem=new STreeItem(item->iTreeView, &file);																

		item->iTreeView->AddUnder(subItem, item);

		// this is so we can double-click on a directory in the file-view and have it
		// automagically be expanded & selected in the tree-view
		if(strlen(item->iAutoSelectDir)) {
			if(strcasecmp(item->iAutoSelectDir, file.Name())==0) {
				int32 itemIndex=item->iTreeView->IndexOf(subItem);
				item->iTreeView->Select(itemIndex);
				
				strcpy(item->iAutoSelectDir, "");
			}
		}
			
		item->iTreeView->Window()->Unlock();
	}

	directory.Unset();

	// we don't need this anymore
	strcpy(item->iAutoSelectDir, "");

	// try locking the window one more time so we can sort the new entries
	terminated=false;
	while (item->iTreeView->Window()->LockWithTimeout(10000)==B_TIMED_OUT) {
		get_sem_info(item->iScanThreadTerminate, &semInfo);
		if(semInfo.count <=0) {
			terminated=true;
			break;
		}
	}
	if(!terminated) {
		// sorting throws away the current selection, so...
		int32 selectIndx=item->iTreeView->FullListCurrentSelection();
		STreeItem *selectItem=(STreeItem *) item->iTreeView->FullListItemAt(selectIndx);	
	
		item->iTreeView->SortItems();
		
		int32 visualIndex=item->iTreeView->IndexOf(selectItem);
		item->iTreeView->Select(visualIndex);
		
		item->iTreeView->Window()->Unlock();
	}
	
	if(terminated) {
		// reset scan flag
		item->iTreeScan=false;
	}
	
	// lower flag
	release_sem(item->iScanThreadActive);
	
	if(terminated) {
		return -1;
	} else {
		return 0;
	}	
}

// request scan termination and wait for it to stop
void STreeItem::RequestScanTerminate()
{
	acquire_sem(iScanThreadTerminate);
		
	// sit and wait for the thread to end
	acquire_sem(iScanThreadActive);
	release_sem(iScanThreadActive);
		
	release_sem(iScanThreadTerminate);
}

STreeView::STreeView(BRect bounds, char *name)
 : ColumnListView(	bounds, 
 					&iContainerView, 
 					name,
 					B_FOLLOW_ALL,				// resizing mode
					B_WILL_DRAW | 	
					B_FRAME_EVENTS |			// flags		 
					B_NAVIGABLE,
					B_SINGLE_SELECTION_LIST,	// view type
					true,						// hierarchical	*/
					true,						// horizontal scrollbars
					true,						// vertical scrollbars
					true, 						// scroll-view corner
					B_FANCY_BORDER) 			// border style
{
	// build columns
	this->BuildColumns();
	
	// set sorting function
	SetSortFunction(Compare);

	volumeRoster=new BVolumeRoster();
}				

STreeView::~STreeView()
{
	delete volumeRoster;
}

void STreeView::MouseDown(BPoint point)
{
	BPoint tPoint;
	uint32 buttons;

	int32 itemIndex=IndexOf(point);

	// get mouse buttons
	GetMouse(&tPoint, &buttons, false);

	// right mouse button (only) ?
	if(itemIndex && (buttons & B_SECONDARY_MOUSE_BUTTON)) {

		// grab focus
		MakeFocus();
	
		STreeItem *item=(STreeItem *) ItemAt(itemIndex);
	
		// select the item
		Select(itemIndex);
	
		// let the main window handle the menus
		BMessage *message;
		if(item->File()->IsTrash()) {
			message=new BMessage(MSG_PIONEER_TRASH_POPUP_MENU);
		} else if(item->IsVolume()) {
			message=new BMessage(MSG_PIONEER_VOLUME_POPUP_MENU);
		} else {
			message=new BMessage(MSG_PIONEER_FOLDER_POPUP_MENU);
		}
		message->AddPoint("point", ConvertToScreen(point));
		message->AddPointer("item", item);
		message->AddInt8("fileview",0);
		
		Window()->Looper()->PostMessage(message);
		delete message;

	} else {
		ColumnListView::MouseDown(point);
	}
}

void STreeView::KeyDown(const char *bytes,int32 numbytes)
{
	if(numbytes==1)
	{
		switch(bytes[0])
		{
			// contract item or move up to parent item if contracted
			case B_LEFT_ARROW:
			{
				CLVListItem *item=(CLVListItem*)ItemAt(CurrentSelection());
				if(!item)
					break;
					
				if(item->IsExpanded())
					Collapse(item);
				else
				{
					// This is a bit of a hack because SuperItem doesn't
					// seem to work for ColumnListViews
					int32 index=FullListIndexOf(item);
					Select(index-1);
				}
				break;
			}
			// expand item or move to first child item if expanded and one exists
			case B_RIGHT_ARROW:
			{
				CLVListItem *item=(CLVListItem*)ItemAt(CurrentSelection());
				if(!item)
					break;
					
				if(item->IsExpanded())
				{
					// This is a bit of a hack because AddItemUnder doesn't
					// seem to work for ColumnListViews
					int32 index=FullListIndexOf(item);
					
					Select(index+1);
				}
				else
					Expand(item);
				break;
			}
			
			// Toggle expansion on these keys
			case B_SPACE:
			case B_ENTER:
			{
				CLVListItem *item=(CLVListItem*)ItemAt(CurrentSelection());
				if(!item)
					break;
					
				if(item->IsExpanded())
					Collapse(item);
				else
					Expand(item);
				break;
			}
			default:
				ColumnListView::KeyDown(bytes,numbytes);
		}
	}
	else
	{
		// Call the inherited version for everything that we don't specifically support
		ColumnListView::KeyDown(bytes,numbytes);
	}
}

void STreeView::BuildColumns() 
{
	// create expander column
	AddColumn(new CLVColumn(NULL,						// label 
							  20.0,						// width
							  CLV_EXPANDER | 			
							  CLV_LOCK_AT_BEGINNING | 	// flags
							  CLV_NOT_MOVABLE));
	
	// create icon column						  
	AddColumn(new CLVColumn(NULL,
							  20.0,
							  CLV_LOCK_AT_BEGINNING |
							  CLV_NOT_MOVABLE | 
							  CLV_NOT_RESIZABLE | 
							  CLV_PUSH_PASS | 
							  CLV_MERGE_WITH_RIGHT));
	
	// create folder name column						  
	AddColumn(new CLVColumn("Folder",
							  200.0,
					  	  	  CLV_SORT_KEYABLE,
							  50.0));
}

// sorting function
int STreeView::Compare(const CLVListItem* aItem1, const CLVListItem* aItem2, int32 sortKey)
{
	STreeItem *item1=(STreeItem *) aItem1;
	STreeItem *item2=(STreeItem *) aItem2;

	// sort names
	return strcasecmp(item1->File()->Name(), item2->File()->Name());
}

// refresh tree
void STreeView::Refresh()
{
	// save entries that are expanded
	// ...
	
	// delete old items
	STreeItem *oldItem;
	for (uint32 i=0; (oldItem=(STreeItem *) FullListItemAt(i)); i++) {
		delete oldItem;
	}
			
	// clear list
	MakeEmpty();

	// add Desktop root
	BEntry entry(DESKTOP_ENTRY);
	SFile file(&entry); 
	STreeItem *desktopItem=new STreeItem(this, &file, false);
    AddItem(desktopItem);
    entry.Unset();
 
    // add volumes
	BVolume volume;
   	char volName[B_FILE_NAME_LENGTH];
	
    while(volumeRoster->GetNextVolume(&volume)==B_NO_ERROR)
    {
    	// skipped unmounted volumes
    	volume.GetName(volName);
    	if(strlen(volName)==0)
    		continue;
    
    	SFile volRoot(&volume);

    	STreeItem *volumeItem=new STreeItem(this, &volRoot);
    	AddUnder(volumeItem, desktopItem);
    }
    
    // select and expand desktop entry on refresh
    MakeFocus();
    Select(0);
    Expand(desktopItem);

    // re-expand previous selections
    // ...
}

// current selections
void STreeView::CurrentSelections(SFileList *fileList)
{
	if(!fileList) {
		return;
	}

	for (int32 i=0; i < CountItems(); i++) {
		STreeItem *item=(STreeItem *) ItemAt(i); 
		if(item->IsSelected()) {
			fileList->AddItem((void *) item->File());
		}		
	}
}

// expand and select
void STreeView::ExpandAndSelect(BMessage *message)
{
	if(message==NULL) {
		return;
	}

	entry_ref entryRef;
	if(message->FindRef("dir", 0, &entryRef) !=B_OK) {
		return;
	}

	STreeItem* treeItem;
	if(message->FindPointer("tree_item", 0, (void **) &treeItem) !=B_OK) {
		return;
	}

	SFile file(&entryRef);
	if(treeItem) treeItem->ExpandAndSelect(&file);
}

// node monitor: scan for created
void STreeView::ScanForCreated(const SFile *file)
{
	entry_ref entryRef=*file->Ref();

	// find the parent of the entry
	BEntry childEntry(&entryRef);
	
	// make sure the is a directory (or symbolic link to a possible directory)
	if(!childEntry.IsDirectory() && !childEntry.IsSymLink()) {
		return;
	}
	
	BEntry parentEntry;
	childEntry.GetParent(&parentEntry);
	
	entry_ref parentEntryRef;
	parentEntry.GetRef(&parentEntryRef);
	
	// find the parent in the tree by comparing entry_ref's
	STreeItem *item;
	bool found=false;
	
	for (uint32 i=0; (item=(STreeItem *) FullListItemAt(i)); i++) {
	
		// found it ?
		if(*item->File()->Ref()==parentEntryRef) {
			found=true;
			break;
		}
	}
		
	if(!found) {
		// I'm sorry, there must be some kind of mistake...
		return;
	}
	
	// if we haven't scanned it yet, why bother ?
	if(!item->IsScanned()) {
		return;
	}

	// check to see if we've already added it to the list
	// this prevents duplicate entries from being created
	// when we rename a folder
	STreeItem *existItem;
	found=false;

	for (uint32 i=0; (existItem=(STreeItem *) FullListItemAt(i)); i++) {
	
		// found it ?
		if(strcasecmp(existItem->File()->Name(), entryRef.name)==0) {
			found=true;
			break;
		}
	}

	if(found) {
		return;
	}

	// add the entry
	STreeItem *subItem=new STreeItem(this, file);																
	AddUnder(subItem, item);
	
	SortItems();
}

// node monitor: scan for moved
void STreeView::ScanForMoved(const SFile *fromFile, const SFile *toFile)
{
	// use the standard creation routine
	ScanForCreated(toFile);
	ScanAllForRemoved();
}

// node monitor: scan for removed
void STreeView::ScanAllForRemoved()
{
	STreeItem *item;
	for (uint32 i=0; (item=(STreeItem *) FullListItemAt(i)); i++)
	{
		if(!item->File()->Exists())
		{
			// refresh the file view if the item is selected
			if(item->IsSelected())
			{
				BMessage *message=new BMessage(MSG_FILEVIEW_CLEAR);
				Looper()->PostMessage(message);
				delete message;
			}
		
			// remove it !
			RemoveItem(item);
			
			delete item;
		}	
	}
}

// node monitor: update trash icon
void STreeView::ScanUpdateTrashIcon()
{
	// find the 'Trash' folder in the first level
	CLVListItem *desktopItem=FullListItemAt(0);
	for (int32 i=0; i < FullListNumberOfSubitems(desktopItem); i++) {
		STreeItem *subItem=(STreeItem *) FullListItemAt(i);
		if(subItem->File()->IsTrash()) {
			subItem->ResetIcon();
		}

	}
}

void STreeView::SpawnTracker()
{
	STreeItem *item=(STreeItem*)ItemAt(CurrentSelection());

	if(!item)
		return;

	BString string("/boot/beos/system/Tracker '");
	string+=item->File()->PathDesc();
	string+="' &";
	system(string.String());
}
