#include <stdlib.h>

#include <ScrollBar.h>

#include "Colors.h"

#include "SplitterView.h"

typedef struct {
	BView *view;
	BView *sibling;
	uint32 pane;
} SplitterViewChild;

CSplitterView::CSplitterView(BRect frame,
					   		  const char *name,
					   		  uint32 resizingMode,
					   		  uint32 flags)
 : BView(frame, name, resizingMode, flags),
   iTracking(false),
   iSplitFactor(0.30),
   iSplitLineWidth(4.0),
   iChildList(0)
{
	/* create internal list */
	iChildList=new BList();

	/* set color of split bar */
	rgb_color splitColor;
	splitColor.red=203; splitColor.blue=203; splitColor.green=203; splitColor.alpha=255;
	SetHighColor(splitColor);
}	

/* destructor */
CSplitterView::~CSplitterView()
{
	if(iChildList) {
		SplitterViewChild *child;
		for (int32 i=0; (child=(SplitterViewChild *) iChildList->ItemAt(i)); i++) {
			delete child;
		}
	
		delete iChildList;
	}
}

/* add a child */
void CSplitterView::AddChild(BView *aView, 
					   		  uint32 pane,	
					    	  BView *sibling)
{
	/* add to internal list */
	SplitterViewChild *child=new SplitterViewChild();
	child->view=aView;
	child->sibling=sibling;
	child->pane=pane;
	iChildList->AddItem(child);
	
	/* recalculate bounds */
	CalcChildBounds(aView, pane);

	BView::AddChild(aView, sibling);
}

/* remove a child */
bool CSplitterView::RemoveChild(BView *aView)
{
	/* remote from internal list */
	SplitterViewChild *child;
	for (int32 i=0; (child=(SplitterViewChild *) iChildList->ItemAt(i)); i++) {
		if(child->view==aView) {
			iChildList->RemoveItem(child);
			delete child;
		}
	}			

	return BView::RemoveChild(aView);
}

/* implemented Draw function */
void CSplitterView::Draw(BRect rect)
{
	/* draw splitter line */
	iSplitLineRect.top=Bounds().top;
	iSplitLineRect.bottom=Bounds().bottom;
	iSplitLineRect.left=((Bounds().right - Bounds().left) * iSplitFactor) - (iSplitLineWidth/2.0);
	iSplitLineRect.right=((Bounds().right - Bounds().left) * iSplitFactor) + (iSplitLineWidth/2.0);
	
	FillRect(iSplitLineRect);
	
	/* give it a more 3D look -- oooh, aaah */ 
	SetHighColor(BeHighlight);
	StrokeLine(BPoint(iSplitLineRect.left, iSplitLineRect.top), 
				BPoint(iSplitLineRect.left, iSplitLineRect.bottom));
				
	SetHighColor(BeDarkShadow);
	StrokeLine(BPoint(iSplitLineRect.right, iSplitLineRect.top), 
				BPoint(iSplitLineRect.right, iSplitLineRect.bottom));

}

/* implemented MouseMoved function */
void CSplitterView::MouseMoved(BPoint point, uint32 transit, const BMessage *message)
{
	if(!iTracking) {
		return;
	}

	/* recalculate split factor */
	iSplitFactor=point.x / Bounds().Width();	

	/* recalculate bounds of children */
	SplitterViewChild *child;
	for (int32 i=0; (child=(SplitterViewChild *) iChildList->ItemAt(i)); i++) {
		CalcChildBounds(child->view, child->pane);
	}
	
	Invalidate();
}

/* implement MouseDown function */
void CSplitterView::MouseDown(BPoint point)
{
	if(iTracking) {
		return;
	}
	
	if(iSplitLineRect.Contains(point)) {
		iTracking=true;
		SetMouseEventMask(B_POINTER_EVENTS, B_LOCK_WINDOW_FOCUS | B_NO_POINTER_HISTORY);
	}
}

/* implement MouseUp function */
void CSplitterView::MouseUp(BPoint point)
{
	if(!iTracking) {
		return;
	}

	iTracking=false;

	/* prevent the user from dragging the splitbar out of bounds */
	if(point.x < iSplitLineWidth) {
		point.x=(iSplitLineWidth / 2.0);
	} else if(point.x > (Bounds().Width() - (iSplitLineWidth / 2.0))) {
		point.x=(Bounds().Width() - (iSplitLineWidth / 2.0));
	}

	/* recalculate split factor */
	iSplitFactor=point.x / Bounds().Width();	
	
	/* recalculate bounds of children */
	SplitterViewChild *child;
	for (int32 i=0; (child=(SplitterViewChild *) iChildList->ItemAt(i)); i++) {
		CalcChildBounds(child->view, child->pane);
	}
	
	Invalidate();
}

/* implement FrameResized function */
void CSplitterView::FrameResized(float width, float height)
{
	/* recalculate bounds of children */
	SplitterViewChild *child;
	for (int32 i=0; (child=(SplitterViewChild *) iChildList->ItemAt(i)); i++) {
		CalcChildBounds(child->view, child->pane);
	}
	
	Invalidate();
}

/* pre-calculate bounds for a particular pane */
BRect CSplitterView::CalcPaneBounds(uint32 pane)
{
	BRect rect;

	rect.top=Bounds().top;
	rect.bottom=Bounds().bottom;
	if(pane==C_LEFT) {
		rect.left=Bounds().left;
		rect.right=((Bounds().right - Bounds().left) * iSplitFactor) - (iSplitLineWidth/2.0) - 1.0;
	} else {
		rect.left=((Bounds().right - Bounds().left) * iSplitFactor) + (iSplitLineWidth/2.0) + 1.0;
		rect.right=Bounds().right;
	}
	
	return rect;
}

/* calculate child bounds */
BRect CSplitterView::CalcChildBounds(BView *aView, uint32 pane)
{
	BRect rect=aView->Bounds();
	if(aView->ResizingMode() & B_FOLLOW_TOP) {
		rect.top=Bounds().top;
	}
	if(aView->ResizingMode() & B_FOLLOW_BOTTOM) {
		rect.bottom=Bounds().bottom;
	}
	if(pane==C_LEFT) {
		if(aView->ResizingMode() & B_FOLLOW_LEFT) {
			rect.left=Bounds().left;
		}
		if(aView->ResizingMode() & B_FOLLOW_RIGHT) {
			rect.right=((Bounds().right - Bounds().left) * iSplitFactor) - (iSplitLineWidth/2.0) - 1.0;
		}
	} else {
		if(aView->ResizingMode() & B_FOLLOW_LEFT) {
			rect.left=((Bounds().right - Bounds().left) * iSplitFactor) + (iSplitLineWidth/2.0) + 1.0;
		}
		if(aView->ResizingMode() & B_FOLLOW_RIGHT) {
			rect.right=Bounds().right;
		}
	}
	
	aView->MoveTo(rect.left, rect.top);
	aView->ResizeTo(rect.Width(), rect.Height());

	return rect;
}

